/*
 * p7array_impl.h -- ParentheTcl (p7tcl)
 *
 # Copyright (c) 2004-2006 Henry Strickland
 #
 # Permission is hereby granted, free of charge, to any person obtaining a
 # copy of this software and associated documentation files (the "Software"),
 # to deal in the Software without restriction, including without limitation
 # the rights to use, copy, modify, merge, publish, distribute, sublicense,
 # and/or sell copies of the Software, and to permit persons to whom the
 # Software is furnished to do so, subject to the following conditions:
 #
 # The above copyright notice and this permission notice shall be included
 # in all copies or substantial portions of the Software.
 #
 # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 # THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 # OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 # ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 # OTHER DEALINGS IN THE SOFTWARE.
 #
 #         (* http://www.opensource.org/licenses/mit-license.php *)
 *
 */

void FUNC(Link)( ARRAY p ) {
    if (!p) return;
#ifdef DEBUG
fdebugf(stderr, "LINK %lx %d\n", (ulong)p, p->ref_count );
#endif
#ifdef P7_MAGIC
    assert( p->magic == P7_MAGIC );
#endif
CHECK
    ++ p->ref_count;
}

void FUNC(Free)( ARRAY p ) {
    if (!p) return;
#ifdef DEBUG
fdebugf(stderr, "FREE %lx %d [%d]*%d\n", (ulong)p, p->ref_count, p->length, sizeof(NAME) );
#endif
#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
#endif
	if ( p->cls && p->cls->cleanup ) {
	    p->cls->cleanup ( (p7obj)p );
	}
#ifdef P7_MAGIC
	p->magic= 0xDeadBeef;
#endif
CHECK
	/* possibly destroy all data with ~~~~ */
	/* memset( p, '~', sizeof (*p) + p->length * sizeof(NAME) ); */
	p7free( "ARRAY", (void*)p );
CHECK
}

void FUNC(Unlink)( ARRAY p ) {
    if (!p) return;
#ifdef DEBUG
fdebugf(stderr, "UNLINK %lx %d\n", (ulong)p, p->ref_count );
#endif
#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
#endif
CHECK
	-- p->ref_count;
	assert( p->ref_count >= 0 );
	if (p->ref_count == 0) {
		FUNC(Free)(p);
	}
CHECK
}

p7err FUNC(Assign)( p7world world, ARRAY* rv, ARRAY x) {
#ifdef DEBUG
fdebugf(stderr, "ASSIGN @%lx: %lx <- %lx\n", (ulong)rv, (ulong)(*rv), (ulong)x );
#endif
#ifdef P7_MAGIC
	if (x) assert( x->magic == P7_MAGIC );
	if (*rv) assert( (*rv)->magic == P7_MAGIC );
#endif
	/* Link before Unlink -- in case they are the same, dont let it free! */
	if (x) p7obj_Link( (p7obj)x);
	if (*rv) p7obj_Unlink( (p7obj)*rv); 
	(*rv)= x;
CHECK
	return TCL_OK;
}

p7err FUNC(New)( p7world world, ARRAY* rv, int n) {
	    
CHECK
	if (n<0) return p7set_error_static( world, "negative length in New()" );
CHECK

	if (*rv) p7obj_Unlink( (p7obj)*rv); 

	if ( n==0 ) {
	    /* EMPTY array is the NULL array. */
	    (*rv)= NULL;
	    return TCL_OK;
	}

	(*rv)= (ARRAY)p7malloc( "ARRAY", sizeof (**rv) + n*sizeof(NAME) );
CHECK
	if (!(*rv)) return p7set_error_static( world, "Out Of Memory" );

#ifdef DEBUG
fdebugf(stderr, "NEW %lx [%d]\n", (ulong)(*rv), n );
#endif
#ifdef P7_MAGIC
	(*rv)->magic= P7_MAGIC;
#endif
	(*rv)->ref_count= 0;
	(*rv)->cls= & CLS;
	(*rv)->length= n;
CHECK
	return TCL_OK;
}

p7err FUNC(Subscript)( p7world world, NAME **rv, ARRAY *pRef, int i ) {
	ARRAY p= *pRef;
CHECK
	if (!p) return p7set_error_static(world, "Subscript() on EMPTY array");
#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
#endif
	if ( i<0 ) {
		return p7set_error_static(world, "bounds error: index is negative");
	}
	if ( i>=p->length ) {
		return p7set_error_static(world, "bounds error: index too big");
	}
	(*rv)= p->data+i;
CHECK
	return TCL_OK;
}

p7err FUNC(GetAt)( p7world world, NAME *rv, ARRAY p, int i ) {
CHECK
	if (!p) return p7set_error_static(world, "GetAt() on EMPTY array");
#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
#endif
	if ( i<0 || i>=p->length ) {
		return p7set_error_static(world, "bounds error in GetAt()");
	}
	(*rv)= p->data[i];
CHECK
	return TCL_OK;
}


p7err FUNC(PutAt)( p7world world, ARRAY p, int i, NAME x ) {
CHECK
	if (!p) return p7set_error_static(world, "PutAt() on EMPTY array");
#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
#endif
CHECK
	if ( i<0 || i>=p->length ) {
		return p7set_error_static(world, "bounds error in PutAt()");
	}
	p->data[i]= x;
CHECK
	return TCL_OK;
}


