/*
 * parenthetcl.h -- ParentheTcl (p7tcl)
 *
 # Copyright (c) 2004-2006 Henry Strickland
 #
 # Permission is hereby granted, free of charge, to any person obtaining a
 # copy of this software and associated documentation files (the "Software"),
 # to deal in the Software without restriction, including without limitation
 # the rights to use, copy, modify, merge, publish, distribute, sublicense,
 # and/or sell copies of the Software, and to permit persons to whom the
 # Software is furnished to do so, subject to the following conditions:
 #
 # The above copyright notice and this permission notice shall be included
 # in all copies or substantial portions of the Software.
 #
 # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 # THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 # OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 # ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 # OTHER DEALINGS IN THE SOFTWARE.
 #
 #         (* http://www.opensource.org/licenses/mit-license.php *)
 *
 */

#ifdef DEBUG
#define TCL_MEM_DEBUG 1
#endif

#include <malloc.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <tcl.h>

/* our route to memory, thru p7base.c */
void* p7malloc_func( char* what, char* file, int line, size_t size );
void p7free_func( char* what, char* file, int line, void* ptr );
void p7check( char* file, int line );

/* wrap the route to memory with FILE and LINE */
#define p7malloc(W,S) p7malloc_func(W,__FILE__,__LINE__,S)
#define p7free(W,P) p7free_func(W,__FILE__,__LINE__,P)

#ifdef TCL_MEM_DEBUG

#define CHECK p7check(__FILE__,__LINE__);

#else /*TCL_MEM_DEBUG*/

#define CHECK /* no check */

#endif /*TCL_MEM_DEBUG*/

#if DEBUG>90

#define ATLINE(N) fprintf(stderr, ";<%d>;", N );
#define DOING(S) fprintf(stderr, ";<%s>;", S );
#define fdebugf fprintf

#else 

#define ATLINE(N) 
#define DOING(S) 
#define fdebugf if(0)fprintf

#endif 


#ifdef DEBUG
#define P7_WORLDMAGIC 123056789
#define P7_MAGIC 987650321
//#undef P7_WORLDMAGIC
//#undef P7_MAGIC
#else
#undef P7_WORLDMAGIC
#undef P7_MAGIC
#endif


typedef int p7err;		 // TCL_OK or TCL_ERROR 
typedef int p7int;
typedef unsigned int p7uint;      // only used in >>>
typedef double p7double;

typedef int p7zero; /* for multipurpose constant 0 */

// to help overloading in C++, p7byt is unsigned, but p7utf is signed.
typedef unsigned char p7byt;	// must be unsigned
typedef signed char p7utf;	// must be signed
typedef unsigned short p7uni;	// unsigned, as in java

typedef Tcl_Obj* p7tcl; /* TCL_OK or TCL_ERROR */

typedef struct struct_p7world {
#ifdef P7_WORLDMAGIC
    int	    magic;
#endif
    int	    ref_count;
    Tcl_Interp	    * interp;
    struct struct_p7globals* globals;
    int	    num_literals;
    p7tcl   * literals;  /* array of num_literals of p7tcl */
} *p7world;


typedef struct struct_p7obj {
#ifdef P7_MAGIC
    int	    magic;
#endif
    int	    ref_count;
    struct struct_p7cls*   cls;
} *p7obj; /* either array or p7class */

typedef struct struct_p7cls {
#ifdef P7_MAGIC
    int	    magic;
#endif
    int	    ref_count;
    struct struct_p7cls*   cls;

    char*   clsname;
    int	    clsnum;
    int	    clssize;
    void    (*cleanup)(p7obj);
    int			   vectorRank;  // number of dimensions
    struct struct_p7cls*   elementCls;
} *p7cls;



extern struct Tcl_ObjType p7classObjType;

extern struct struct_p7cls p7class_p7dict;

extern struct struct_p7cls p7class_p7uni_VEC_;
extern struct struct_p7cls p7class_p7utf_VEC_;
extern struct struct_p7cls p7class_p7byt_VEC_;
extern struct struct_p7cls p7class_p7int_VEC_;
extern struct struct_p7cls p7class_p7double_VEC_;
extern struct struct_p7cls p7class_p7tcl_VEC_;
extern struct struct_p7cls p7class_p7obj_VEC_;
extern struct struct_p7cls p7class_p7dict_VEC_;


#include "p7array.h"

typedef struct struct_p7dict {
#ifdef P7_MAGIC
    int	    magic;
#endif
    int	    ref_count;
    p7cls	cls;

    Tcl_HashTable  ht;
    int		count;  // number of entries
} *p7dict;


    		

#define p7int_Assign(world,RV,X) ( *(RV) = (p7int)(X) )
#define p7double_Assign(world,RV,X) ( *(RV) = (p7double)(X) )
#define p7byt_Assign(world,RV,X) ( *(RV) = (p7byt)(X) )
#define p7utf_Assign(world,RV,X) ( *(RV) = (p7utf)(X) )
#define p7uni_Assign(world,RV,X) ( *(RV) = (p7uni)(X) )

void p7obj_VEC__Cleanup( p7obj obj );

p7err p7new_object_vector( p7world world, p7obj_VEC_* rv, int len, p7cls cls );


/* set_error_static sets interp result to static what, and returns TCL_ERROR */
p7err p7set_error_static( p7world world, char* what );
p7err p7set_error_d( p7world world, char* what, int d );
p7err p7set_error_s( p7world world, char* what, char* s );

p7err p7cmp( p7tcl a, p7tcl b );
p7err p7eq( p7tcl a, p7tcl b );
p7err p7ne( p7tcl a, p7tcl b );
p7err p7lt( p7tcl a, p7tcl b );
p7err p7le( p7tcl a, p7tcl b );
p7err p7gt( p7tcl a, p7tcl b );
p7err p7ge( p7tcl a, p7tcl b );

p7err p7tclcond(p7world world, int* rv, p7tcl tcl);

/* int */
p7err p7tcl_to_p7int( p7world, p7int* rv, Tcl_Obj* x );
p7err p7tcl_from_p7int( p7world world, p7tcl *rv, p7int x);
p7err p7tcl_to_p7double( p7world, p7double* rv, Tcl_Obj* x );
p7err p7tcl_from_p7double( p7world world, p7tcl *rv, p7double x);
p7err p7tcl_to_p7uni( p7world, p7uni* rv, Tcl_Obj* x );
p7err p7tcl_from_p7uni( p7world world, p7tcl *rv, p7uni x);
p7err p7tcl_to_p7utf( p7world, p7utf* rv, Tcl_Obj* x );
p7err p7tcl_from_p7utf( p7world world, p7tcl *rv, p7utf x);
p7err p7tcl_to_p7byt( p7world, p7byt* rv, Tcl_Obj* x );
p7err p7tcl_from_p7byt( p7world world, p7tcl *rv, p7byt x);

/* p7obj_array */
p7err p7tcl_to_p7obj_VEC_( p7world, p7obj_VEC_* rv, Tcl_Obj* x, p7cls cls );
p7err p7tcl_from_p7obj_VEC_( p7world world, p7tcl *rv, p7obj_VEC_ x);

/* p7int_array */
p7err p7tcl_to_p7int_VEC_( p7world, p7int_VEC_* rv, Tcl_Obj* x );
p7err p7tcl_from_p7int_VEC_( p7world world, p7tcl *rv, p7int_VEC_ x);

/* p7double_array */
p7err p7tcl_to_p7double_VEC_( p7world, p7double_VEC_* rv, Tcl_Obj* x );
p7err p7tcl_from_p7double_VEC_( p7world world, p7tcl *rv, p7double_VEC_ x);

/* p7tcl_array */
p7err p7tcl_to_p7tcl_VEC_( p7world, p7tcl_VEC_* rv, Tcl_Obj* x );
p7err p7tcl_from_p7tcl_VEC_( p7world world, p7tcl *rv, p7tcl_VEC_ x);

/* p7uni_array */
p7err p7tcl_to_p7uni_VEC_( p7world, p7uni_VEC_* rv, Tcl_Obj* x );
p7err p7tcl_from_p7uni_VEC_( p7world world, p7tcl *rv, p7uni_VEC_ x);

/* p7utf_array */
p7err p7tcl_to_p7utf_VEC_( p7world, p7utf_VEC_* rv, Tcl_Obj* x );
p7err p7tcl_from_p7utf_VEC_( p7world world, p7tcl *rv, p7utf_VEC_ x);

/* p7byt_array */
p7err p7tcl_to_p7byt_VEC_( p7world, p7byt_VEC_* rv, Tcl_Obj* x );
p7err p7tcl_from_p7byt_VEC_( p7world world, p7tcl *rv, p7byt_VEC_ x);


/* p7tcl */
p7err p7tcl_to_p7tcl( p7world world, p7tcl* rv, Tcl_Obj* x );
p7err p7tcl_from_p7tcl( p7world world, p7tcl *rv, p7tcl x);
void p7tcl_Link( p7tcl x);
void p7tcl_Unlink( p7tcl x);

p7err p7tcl_to_p7obj( p7world world, p7obj* rv, p7tcl tcl, p7cls cls );
p7err p7tcl_from_p7obj( p7world world, p7tcl *rv, p7obj p);




p7err p7tcl_Assign( p7world world ,p7tcl *rv, p7tcl x );
p7err p7tcl_Subscript( p7world world ,p7tcl **rv, p7tcl *aRef, int i );


p7err p7obj_Assign( p7world world, p7obj* rv, p7obj x) ;
p7err p7obj_New( p7world world, p7obj* rv, p7cls cls ) ;
void p7obj_Link( p7obj p ) ;
void p7obj_Free( p7obj p ) ;
void p7obj_Unlink( p7obj p ) ;

p7tcl p7tcl_newifnull ( p7tcl x );
p7tcl p7tcl_tclpeek ( p7tcl x );
p7tcl p7tcl_p7peek ( p7obj x );


p7int p7unsigned_shift_right( p7int a, p7int b );



p7tcl p7tcl_new_linked(void);

p7err p7initialize( p7world world );

void p7world_link(p7world world);
void p7world_unlink(ClientData cd);
void p7globals_unlink(p7world world);
void p7literals_unlink(p7world world);

p7err p7dict_New(p7world world, p7dict* rv);
p7err p7dict_Subscript(p7world world, p7tcl **rv, p7dict d, p7tcl key);
p7err p7dict_UnsetKey(p7world world, p7dict d, p7tcl key);
p7err p7dict_Assign( p7world world, p7dict* rv, p7dict a); 

#define p7dict_Link(p)		p7obj_Link(((p7obj)p))
#define p7dict_Unlink(p)	p7obj_Unlink(((p7obj)p))

#ifndef DEBUG
#define p7tcl_Link(x)	{if (x) Tcl_IncrRefCount(x);}
#define p7tcl_Unlink(x)	{if (x) Tcl_DecrRefCount(x);}

#define p7obj_Link(p)	{if (p) ++ (p)->ref_count;}
#define p7obj_Unlink(p)	{if (p) {-- (p)->ref_count; if (!(p)->ref_count) p7obj_Free(p); }}
#endif

#ifdef NEVER_FREE
#undef p7obj_Link(p)
#undef p7obj_Unlink(p)

#define p7obj_Link(p)	{ 0; }
#define p7obj_Unlink(p)	{ 0; }
#endif



